"""
One-dimensional grid task with noise

In this task, the agent has the challenge of discriminating between
actual informative state sensors, and a comparatively large number
of sensors that are pure noise distractors. Many learning methods
make the implicit assumption that all sensors are informative.
This task is intended to break them.
"""
from __future__ import print_function
import numpy as np

import becca.connector
from becca_test.base_world import World as BaseWorld


class World(BaseWorld):
    """
    One-dimensional grid world with noise.

    In this world, the agent steps forward and backward
    along three positions on a line. The second position is rewarded
    and the first and third positions are punished. Also, any actions
    are penalized somewhat. It also includes some inputs that are pure noise.
    Optimal performance is a reward of about .70 per time step.

    Most of this world's attributes are defined in base_world.py.
    The few that aren't are defined below.
    """
    def __init__(self, lifespan=None):
        """
        Set up the world.
        """
        BaseWorld.__init__(self, lifespan)
        self.name = 'grid_1D_noise'
        self.name_long = 'noisy one dimensional grid world'
        print("Entering", self.name_long)
        self.num_real_sensors = 3
        # num_noise_sensors : int
        #     Of the sensors, these are purely noise.
        #     These have no basis in the world and are only meant to distract.
        # num_real_sensors : int
        #     Of the sensors, these are the ones that represent position.
        self.num_noise_sensors = 0
        self.num_sensors = self.num_noise_sensors + self.num_real_sensors
        self.num_actions = 2
        self.action = np.zeros(self.num_actions)
        # energy_cost : float
        #     The punishment per position step taken.
        self.energy_cost = 0.01
        # jump_fraction : float
        #     The fraction of time steps on which the agent jumps to
        #     a random position.
        self.jump_fraction = 0.1

        # world_state : float
        #     The actual position of the agent in the world.
        #     This can be fractional.
        self.world_state = 0
        # simple_state : int
        #     The nearest integer position of the agent in the world.
        self.simple_state = 0

        self.world_visualize_period = 1e6
        self.brain_visualize_period = 1e3


    def step(self, action):
        """
        Take one time step through the world

        Parameters
        ----------
        action : array of floats
            The set of action commands to execute.

        Returns
        -------
        reward : float
            The amount of reward or punishment given by the world.
        sensors : array of floats
            The values of each of the sensors.
        """
        self.action = action.copy().ravel()
        self.action[np.nonzero(self.action)] = 1.
        self.timestep += 1
        step_size = self.action[0] - self.action[1]

        # An approximation of metabolic energy.
        energy = self.action[0] + self.action[1]
        self.world_state = self.world_state + step_size

        # At random intervals, jump to a random position in the world.
        if np.random.random_sample() < self.jump_fraction:
            self.world_state = (self.num_real_sensors *
                                np.random.random_sample())

        # Ensure that the world state falls between 0 and num_real_sensors.
        self.world_state -= (self.num_real_sensors *
                             np.floor_divide(self.world_state,
                                             self.num_real_sensors))
        self.simple_state = int(np.floor(self.world_state))

        # Assign sensors as zeros or ones.
        # Represent the presence or absence of the current position in the bin.
        real_sensors = np.zeros(self.num_real_sensors)
        real_sensors[self.simple_state] = 1

        # Generate a set of noise sensors
        noise_sensors = np.round(np.random.random_sample(
            self.num_noise_sensors))
        sensors = np.hstack((real_sensors, noise_sensors))
        reward = -1.
        if self.simple_state == 1:
            reward = 1.
        reward -= energy * self.energy_cost

        return sensors, reward


    def visualize_world(self, brain):
        """
        Show what's going on in the world.
        """
        state_image = ['.'] * (self.num_real_sensors +
                               self.num_actions + 2)
        state_image[self.simple_state] = 'O'
        state_image[self.num_real_sensors:self.num_real_sensors + 2] = '||'
        action_index = np.where(self.action > 0.1)[0]
        if action_index.size > 0:
            state_image[self.num_real_sensors + 2 + action_index[0]] = 'x'
        print(''.join(state_image))


if __name__ == "__main__":
    becca.connector.run(World())



"""
Two-dimensional grid task.

This task is a 2D extension of the 1D grid world and
is similar to it in many ways. It is a little more
challenging, because can take two actions to reach a reward state.
"""
from __future__ import print_function
import numpy as np

import becca.connector
from becca_test.base_world import World as BaseWorld


class World(BaseWorld):
    """
    Two-dimensional grid world.

    In this world, the agent steps North, South, East, or West in
    a 5 x 5 grid-world. Position (4,4) is rewarded and (2,2)
    is punished. There is also a lesser penalty for each
    horizontal or vertical step taken.
    Optimal performance is a reward of about .9 per time step.

    Some of this world's attributes are defined in base_world.py.
    The others are defined below.
    """
    def __init__(self, lifespan=None):
        """
        Initialize the world.

        Parameters
        ----------
        lifespan : int
            The number of time steps to continue the world.
        """
        BaseWorld.__init__(self, lifespan)
        self.name = 'grid_2D'
        self.name_long = 'two dimensional grid world'
        print("Entering", self.name_long)

        self.num_actions = 8
        # world_size : int
        #     The world consists of a 2D grid of size
        #     world_size by world_size.
        self.world_size = 5
        self.num_sensors = self.world_size ** 2
        # world_state : float
        #     The actual position of the agent in the world.
        #     This can be fractional.
        self.world_state = np.array([1., 1.])
        # targets : list of tuples of ints
        #     Each tuple is a (row, column) pair indicating a location
        #     that is rewarded.
        #     Reward positions (2,2) and (4,4)
        self.targets = [(1, 1), (3, 3)]
        self.action = np.zeros(self.num_actions)
        # energy_cost : float
        #     The punishment per position step taken.
        self.energy_cost = 0.05
        # jump_fraction : float
        #     The fraction of time steps on which the agent jumps to
        #     a random position.
        self.jump_fraction = 0.1
        # obstacles : list of tuples of ints
        #     Each tuple is a (row, column) pair indicating a location
        #     that are punished.
        #     Punish positions (2,4) and (4,2)
        self.obstacles = [(1, 3), (3, 1)]

        self.world_visualize_period = 1e6
        self.brain_visualize_period = 1e3


    def step(self, action):
        """
        Advance the world by one time step.

        Parameters
        ----------
        action : array of floats
            The set of action commands to execute.

        Returns
        -------
        reward : float
            The amount of reward or punishment given by the world.
        sensors : array of floats
            The values of each of the sensors.
        """
        # Turn the action command into a change in the world.
        self.action = action.ravel()
        self.action[np.nonzero(self.action)] = 1.
        self.timestep += 1
        self.world_state += (self.action[0:2] -
                             self.action[4:6] +
                             2 * self.action[2:4] -
                             2 * self.action[6:8]).T
        energy = (np.sum(self.action[0:2]) +
                  np.sum(self.action[4:6]) +
                  np.sum(2 * self.action[2:4]) +
                  np.sum(2 * self.action[6:8]))

        # At random intervals, jump to a random position in the world.
        if np.random.random_sample() < self.jump_fraction:
            self.world_state = (
                np.random.randint(0, self.world_size,
                                  size=len(self.world_state)).astype(float))

        # Enforce lower and upper limits on the grid world
        # by looping them around.
        self.world_state = np.remainder(self.world_state, self.world_size)
        sensors = self.assign_sensors()

        # Assign the reward appropriate to the current state.
        reward = 0.
        for obstacle in self.obstacles:
            if tuple(self.world_state) == obstacle:
                reward = -1.
        for target in self.targets:
            if tuple(self.world_state) == target:
                reward = 1.
        reward -= self.energy_cost * energy

        return sensors, reward


    def assign_sensors(self):
        """
        Construct the sensor array from the state information.

        Returns
        -------
        sensors : list of floats
            The current state of the world, reflected in the sensors.
        """
        sensors = np.zeros(self.num_sensors)
        sensors[int(self.world_state[0] +
                    self.world_state[1] * self.world_size)] = 1
        return sensors


    def visualize_world(self, brain):
        """
        Show the state of the world and the brain.
        """
        print(''.join(['state', str(self.world_state), '  action',
                       str((self.action[0:2] + 2 * self.action[2:4] -
                            self.action[4:6] - 2 * self.action[6:8]).T)]))


if __name__ == "__main__":
    becca.connector.run(World())



"""
Decoupled two-dimensional grid task.

This is just like the regular 2D grid task, except that it the rows
and columns are sensed separately. This makes the task challenging.
Both the row number and column number need to be taken into
account in order to know what actions to take. This task requires building
basic sensory data into more complex features in order to do well.
"""
from __future__ import print_function
import numpy as np

import becca.connector
from becca_test.grid_2D import World as Grid_2D_World


class World(Grid_2D_World):
    """
    Decoupled two-dimensional grid world.

    It's just like the grid_2D world except that the sensors
    array represents a row and a column separately,
    rather than coupled together.

    Optimal performance is a reward of about 90 per time step.

    Attributes
    ----------
    See grid_2D.py for a full description of attributes.
    """
    def __init__(self, lifespan=None):
        """
        Set up the world based on the grid_2D world.

        Parameters
        ----------
        lifespan : int
            The number of time steps to continue the world.
        """
        Grid_2D_World.__init__(self, lifespan)
        self.name = 'grid_2D_dc'
        self.name_long = 'decoupled two dimensional grid world'
        print(", decoupled")
        self.num_sensors = self.world_size * 2


    def assign_sensors(self):
        """
        Create an appropriate sensor array

        Returns
        -------
        sensors : list of floats
            The current state of the world, reflected in the sensors.
        """
        sensors = np.zeros(self.num_sensors)
        # Sensors 0-4 represent each of the 5 rows.
        sensors[int(self.world_state[0])] = 1
        # Sensors 5-9 represent each of the 5 columns.
        sensors[int(self.world_state[1] + self.world_size)] = 1
        return sensors


if __name__ == "__main__":
    becca.connector.run(World())



"""
A multi-step variation on the one-dimensional grid task.

This is intended to be as similar as possible to the
one-dimensional grid task, but requires multi-step planning
or time-delayed reward assignment for optimal behavior.
"""
from __future__ import print_function
import numpy as np

import becca.connector
from becca_test.base_world import World as BaseWorld


class World(BaseWorld):
    """
    One-dimensional grid world, multi-step variation

    In this world, the agent steps forward and backward along a line.
    The fourth position is rewarded and the ninth position is punished.
    Optimal performance is a reward of about 85 per time step.

    Most of this world's attributes are defined in base_world.py.
    The few that aren't are defined below.
    """
    def __init__(self, lifespan=None):
        """
        Initialize the world.

        Parameters
        ----------
        lifespan : int
            The number of time steps to continue the world.
        """
        BaseWorld.__init__(self, lifespan)
        self.name = 'grid_1D_ms'
        self.name_long = 'multi-step one dimensional grid world'
        print("Entering", self.name_long)

        self.num_sensors = 9
        self.num_actions = 2
        self.action = np.zeros(self.num_actions)
        # energy_cost : float
        #     The punishment per position step taken.
        self.energy_cost = 0.01
        # jump_fraction : float
        #     The fraction of time steps on which the agent jumps to
        #     a random position.
        self.jump_fraction = 0.1
        # world_state : float
        #     The actual position of the agent in the world.
        #     This can be fractional.
        self.world_state = 0
        # simple_state : int
        #     The nearest integer position of the agent in the world.
        self.simple_state = 0

        self.world_visualize_period = 1e6
        self.brain_visualize_period = 1e3


    def step(self, action):
        """
        Advance the world by one time step.

        Parameters
        ----------
        action : array of floats
            The set of action commands to execute.

        Returns
        -------
        reward : float
            The amount of reward or punishment given by the world.
        sensors : array of floats
            The values of each of the sensors.
        """
        self.action = action
        self.action = np.round(self.action)
        self.timestep += 1
        energy = self.action[0] + self.action[1]
        self.world_state += self.action[0] - self.action[1]
        # Occasionally add a perturbation to the action to knock it
        # into a different state.
        if np.random.random_sample() < self.jump_fraction:
            self.world_state = self.num_sensors * np.random.random_sample()
        # Ensure that the world state falls between 0 and 9
        self.world_state -= self.num_sensors * np.floor_divide(
            self.world_state, self.num_sensors)
        self.simple_state = int(np.floor(self.world_state))
        if self.simple_state == 9:
            self.simple_state = 0
        # Assign sensors as zeros or ones.
        # Represent the presence or absence of the current position in the bin.
        sensors = np.zeros(self.num_sensors)
        sensors[self.simple_state] = 1
        # Assign reward based on the current state.
        reward = sensors[8] * -1.
        reward += sensors[3]
        # Punish actions just a little.
        reward -= energy * self.energy_cost
        reward = np.max(reward, -1)
        return sensors, reward


    def visualize_world(self, brain):
        """
        Show what's going on in the world.
        """
        state_image = ['.'] * (self.num_sensors + self.num_actions + 2)
        state_image[self.simple_state] = 'O'
        state_image[self.num_sensors:self.num_sensors + 2] = '||'
        action_index = np.where(self.action > 0.1)[0]
        if action_index.size > 0:
            for i in range(action_index.size):
                state_image[self.num_sensors + 2 + action_index[i]] = 'x'
        print(''.join(state_image))


if __name__ == "__main__":
    becca.connector.run(World())



"""
A task in which an robot must choose between ripe and unripe fruit.

In this task, the robot's sensors tell it whether a fruit is either
1) small or large and
2) yellow or purple.
A small yellow fruit is an unripe plum. It is not good to eat.
A small purple fruit is a ripe plum. It is good to eat.
A large yellow fruit is a ripe peach. It is good to eat.
A large purple fruit is a rotten peach. It is not good to eat.

To succeed in this task, the robot has to consider the combination
of its sensors, rather than each one individually. This is
mathematically related to the XOR task, a challenge for
many machine learning algorithms.
"""
from __future__ import print_function
import numpy as np

import becca.connector
from becca_test.base_world import World as BaseWorld


class World(BaseWorld):
    """
    The Fruit selection world.

    In this world, the robot has only two sensors,
    large/small and yellow/purple
    and two actions
    eat or don't eat.

    The robot gets rewarded for eating good fruit and
    punished for eating raw or rotten fruit. This world
    is designed to force the robot to create feature
    from its sensors.

    Most of this world's attributes are defined in base_world.py.
    The few that aren't are defined below.
    """
    def __init__(self, lifespan=None):
        """
        Set up the world.

        Parameters
        ----------
        lifespan : int
            The number of time steps to continue the world.
        """
        BaseWorld.__init__(self, lifespan)
        self.name = 'fruit'
        self.name_long = 'fruit selection world'
        print("Entering", self.name_long)
        self.world_visualize_period = 1e6
        self.brain_visualize_period = 1e3

        # Break out the sensors into
        #    0: large?
        #    1: small?
        #    2: yellow?
        #    3: purple?
        # A sample sensor array would be
        #     [1., 0., 1., 0.]
        # indicating a ripe peach.
        self.num_sensors = 4

        # Break out the actions into
        #     0: eat
        #     1: discard
        self.num_actions = 2
        self.actions = np.zeros(self.num_actions)
        self.reward = 0.

        # acted, eat, discard : boolean
        #     These indicate whether the Becca chose to act on this
        #     time step, and if it did, whether it chose to eat or discard
        #     the fruit it was presented.
        self.acted = False
        self.eat = False
        self.discard = False

        # Grab a piece of fruit to get started.
        self.grab_fruit()


    def grab_fruit(self):
        """
        Grab a new piece of fruit from the box.

        Randomly assign its attributes.
        self.size == 0 # large
        self.size == 1 # small
        self.color == 0 # yellow
        self.color == 1 # purple
        """
        self.size = np.random.randint(2)
        self.color = np.random.randint(2)
        self.edible = ((self.size == 0) and (self.color == 0) or
                       (self.size == 1) and (self.color == 1))

        self.sensors = np.zeros(self.num_sensors)
        if self.size == 0:
            self.sensors[0] = 1.
        if self.size == 1:
            self.sensors[1] = 1.
        if self.color == 0:
            self.sensors[2] = 1.
        if self.color == 1:
            self.sensors[3] = 1.


    def step(self, action):
        """
        Take one time step through the world.

        Parameters
        ----------
        action : array of floats
            The set of action commands to execute.

        Returns
        -------
        self.reward : float
            The amount of reward or punishment given by the world.
        self.sensors : array of floats
            The values of each of the sensors.
        """
        self.timestep += 1
        self.actions = action.ravel()

        # Figure out which action was taken
        self.acted = False
        self.eat = False
        self.discard = False
        if action[0] > .5:
            self.eat = True
            self.acted = True
        elif action[1] > .5:
            self.discard = True
            self.acted = True

        # Check whether the appropriate action was taken, and assign reward.
        # There is a small punishment for doing nothing.
        self.reward = -.1
        if ((self.eat and self.edible) or
                (self.discard and not self.edible)):
            self.reward = 1.
        elif ((self.eat and not self.edible) or
              (self.discard and self.edible)):
            self.reward = -.9

        if self.acted:
            self.grab_fruit()

        return self.sensors, self.reward


    def visualize_world(self, brain):
        """
        Show what's going on in the world.

        Note that this is the reward for the action based on the
        previous set of sensors. This visualizaion is called right after
        the step() method is called, so the sensors have already
        been updated for the next loop.
        """
        state_str = ' || '.join([str(self.sensors),
                                 str(self.actions),
                                 str(self.reward),
                                 str(self.size),
                                 str(self.color),
                                 str(self.timestep)])
        print(state_str)


if __name__ == "__main__":
    becca.connector.run(World())



"""
One-dimensional grid delay task.

This task tests an agent's ability to properly ascribe reward to the
correct cause. The reward is delayed by a variable amount, which
makes the task challenging.
"""
from __future__ import print_function
import numpy as np

import becca.connector
from becca_test.grid_1D import World as Grid_1D_World


class World(Grid_1D_World):
    """
    One-dimensional grid task with delayed reward

    This task is identical to the grid_1D task with the
    exception that reward is randomly delayed a few time steps.

    Most of this world's attributes are defined in base_world.py.
    The few that aren't are defined below.
    """
    def __init__(self, lifespan=None):
        """
        Initialize the world. Base it on the grid_1D world.

        Parameters
        ----------
        lifespan : int
            The number of time steps to continue the world.
        """
        Grid_1D_World.__init__(self, lifespan)
        self.name = 'grid_1D_delay'
        self.name_long = 'one dimensional grid world with delay'
        print('--delayed')

        # max_delay : int
        #     The maximum number of time steps that the reward may be delayed.
        self.max_delay = 1
        # future_reward : list of floats
        #     The reward that has been received, but will not be delivered to
        #     the agent yet. The index of the list indicates how many time
        #     steps will pass before delivery occurs.
        self.future_reward = [0.] * self.max_delay

        self.world_visualize_period = 1e6


    def assign_reward(self, sensors):
        """
        Calcuate the reward corresponding to the current state and assign
        it to a future time step.

        Parameters
        ----------
        sensors : array of floats
            The current sensor values.

        Returns
        -------
        reward : float
            The reward associated the set of input sensors.
        """
        new_reward = -sensors[8]
        new_reward += sensors[3]
        # Punish actions just a little
        new_reward -= self.energy  * self.energy_cost
        # Find the delay for the reward
        delay = np.random.randint(0, self.max_delay)
        self.future_reward[delay] += new_reward
        # Advance the reward future by one time step
        self.future_reward.append(0.)
        reward = self.future_reward.pop(0)
        return reward


    def visualize_world(self, brain):
        """
        Show what's going on in the world.
        """
        state_image = ['.'] * (self.num_sensors + self.num_actions + 2)
        state_image[self.simple_state] = 'O'
        state_image[self.num_sensors:self.num_sensors + 2] = '||'
        action_index = np.where(self.action > 0.1)[0]
        if action_index.size > 0:
            for i in range(action_index.size):
                state_image[self.num_sensors + 2 + action_index[i]] = 'x'
        print(''.join(state_image))


if __name__ == "__main__":
    becca.connector.run(World())



"""
One-dimensional grid task in which the target moves.

This task tests an brain's ability to choose an appropriate action.
It is straightforward. Reward and punishment is clear and immediate.
There is only one reward state and it can be reached in a single
step. The target keeps moving, so it does require the ability to respond
to sensory information.
"""
from __future__ import print_function
import numpy as np

import becca.connector
from becca_test.base_world import World as BaseWorld


class World(BaseWorld):
    """
    One-dimensional grid world with a moving target.

    In this task, the brain steps forward and backward along
    a line. Only one target position is rewarded. Each time
    Becca reaches the target, it jumps to a new position.
    There is also a slight
    punishment for effort expended in taking actions.

    Most of this world's attributes are defined in base_world.py.
    The few that aren't are defined below.
    """
    def __init__(self, lifespan=None):
        """
        Initialize the world.

        Parameters
        ----------
        lifespan : int
            The number of time steps to continue the world.
        """
        BaseWorld.__init__(self, lifespan)
        self.name = 'grid_1D_chase'
        self.name_long = 'one dimensional chase grid world'
        print("Entering", self.name_long)

        # size : int
        #     The number of positions in the 1 dimensional grid.
        self.size = 7
        self.num_sensors = self.size + 2 * (self.size - 1)
        self.num_actions = 2 * (self.size - 1)
        self.reward = 0.
        self.action = np.zeros(self.num_actions)
        # energy : float
        #     The total number of position steps attempted this time step.
        self.energy = 0.
        # energy_cost : float
        #     The punishment per position step taken.
        self.energy_cost = 1e-2
        self.sensors = np.zeros(self.num_sensors)
        # position : int
        #     The position of the agent in the world.
        self.position = 2
        # target_position : int
        #     The position of the target in the world.
        self.target_position = 1

        self.world_visualize_period = 1e6
        self.brain_visualize_period = 1e3


    def step(self, action):
        """
        Advance the world one time step.

        Parameters
        ----------
        action : array of floats
            The set of action commands to execute.

        Returns
        -------
        self.reward : float
            The amount of reward or punishment given by the world.
        self.sensors : array of floats
            The values of each of the sensors.
        """
        self.action = action
        self.action = np.round(self.action)
        self.timestep += 1

        # Find the step size as combinations of the action commands.
        # For world of size 5:
        #     action[i]     result
        #            0      1 step to the right
        #            1      2 steps to the right
        #            2      3 steps to the right
        #            3      4 steps to the right
        #            4      1 step to the left
        #            5      2 steps to the left
        #            6      3 steps to the left
        #            7      4 steps to the left
        scale = np.arange(self.size - 1) + 1.
        step_size = (np.sum(self.action[:self.size - 1] * scale) -
                     np.sum(self.action[self.size - 1:
                                        2 * (self.size - 1)] * scale))
        # Action cost is an approximation of metabolic energy.
        # Action cost is proportional to the number of steps taken.
        self.energy = (np.sum(self.action[:self.size - 1] * scale) +
                       np.sum(self.action[self.size - 1:
                                          2 * (self.size - 1)] * scale))

        self.position += step_size
        self.position = np.minimum(self.position, self.size - 1)
        self.position = int(np.maximum(self.position, 0))
        self.assign_reward()
        self.move_target()

        self.sensors = np.zeros(self.num_sensors)
        # Sense the agent's presence in each bin.
        self.sensors[self.position] = 1
        # Sense the relative distance to the target.
        distance = self.position - self.target_position
        if distance < 0:
            self.sensors[self.size - 1 + np.abs(distance)] = 1
        else:
            self.sensors[2 * (self.size - 1) + np.abs(distance)] = 1

        return self.sensors, self.reward


    def move_target(self):
        """
        Move the target to a new position that is not already occupied.
        """
        while self.target_position == self.position:
            self.target_position = int(np.random.randint(self.size))


    def assign_reward(self):
        """
        Calculate the total reward corresponding to the current state
        """
        self.reward = 0.
        if self.position == self.target_position:
            self.reward += 1.
        # Punish actions just a little
        self.reward -= self.energy  * self.energy_cost


    def visualize_world(self, brain):
        """
        Show what's going on in the world.
        """
        state_image = ['.'] * (self.size + self.num_actions + 2)
        state_image[self.position] = 'O'
        state_image[self.target_position] = '+'
        state_image[self.size:self.size + 2] = '||'
        action_index = np.where(self.action > 0.1)[0]
        if action_index.size > 0:
            for i in range(action_index.size):
                state_image[self.size + 2 + action_index[i]] = 'x'
        state_string = ''.join(state_image)
        print(state_string, '  ', self.timestep, 'time steps')


if __name__ == "__main__":
    becca.connector.run(World())



"""
Two-dimensional visual servo task

Like the 1D visual servo task, this task gives Becca a chance
to build a comparatively large number of sensors into
a few informative features.
"""
from __future__ import print_function
import os

import matplotlib.pyplot as plt
import numpy as np

import becca.connector
from becca_test.base_world import World as BaseWorld
import becca_test.world_tools as wtools


class World(BaseWorld):
    """
    Two-dimensional visual servo world

    In this world, Becca can direct its gaze up, down, left, and
    right, saccading about an image_data of a black square on a white
    background. It is rewarded for directing it near the center.
    Optimal performance is a reward of around .8 reward per time step.


    Some of this world's attributes are defined in base_world.py.
    The rest are defined below.
    """
    def __init__(self, lifespan=None):
        """
        Set up the world.

        Parameters
        ----------
        lifespan : int
            The number of time steps to continue the world.
        """
        BaseWorld.__init__(self, lifespan)
        self.name = 'image_2D'
        self.name_long = 'two dimensional visual world'
        print("Entering", self.name_long)

        # fov_span : int
        #     The world pixelizes its field of view into a superpixel array
        #     that is ``fov_span`` X ``fov_span``.
        self.fov_span = 5
        # Initialize the image_data to be used as the environment.
        module_path = os.path.dirname(os.path.abspath(__file__))
        # image_filename : str
        #     The file name of the image including the relative path.
        self.image_filename = os.path.join(module_path, 'images', 'block_test.png')
        # image_data : array of floats
        #     The image, read in and stored as a 2D numpy array.
        self.image_data = plt.imread(self.image_filename)
        # Convert it to grayscale if it's in color.
        if self.image_data.shape[2] == 3:
            # Collapse the three RGB matrices into one b/w value matrix.
            self.image_data = np.sum(self.image_data, axis=2) / 3.0
        # Define the size of the field of view, its range of
        # allowable positions, and its initial position.
        (im_height, im_width) = self.image_data.shape
        im_size = np.minimum(im_height, im_width)
        # max_step_size : int
        #     The largest step size allowed, in pixels in the original image.
        self.max_step_size = im_size / 2
        # target_column, target_row : int
        #     The row and column index that marks the center of the rewarded region.
        self.target_column = im_width / 2
        self.target_row = im_height / 2
        # reward_region_width : int
        #     The width of the region, in number of columns, within which
        #     the center of the field of view gets rewarded.
        self.reward_region_width = im_size / 8
        # noise_magnitude : float
        #     A scaling factor that drives how much inaccurate each movement
        #     will be.
        self.noise_magnitude = 0.1
        # fov_fraction : float
        #     The approximate fraction of the height and width of the image
        #     that the field of view occupies.
        self.fov_fraction = 0.5
        # fov_height, fov_width : float
        #     The height and width (in number of pixel rows and columns)
        #     of the field of view.
        self.fov_height = im_size * self.fov_fraction
        self.fov_width = self.fov_height
        # column_min, column_max, row_min, row_max : int
        #     The low and high bounds on where the field of view can be centered.
        self.column_min = int(np.ceil(self.fov_width / 2))
        self.column_max = int(np.floor(im_width - self.column_min))
        self.row_min = int(np.ceil(self.fov_height / 2))
        self.row_max = int(np.floor(im_height - self.row_min))
        # column_position, row_position : int
        #     The current location of the center of the field of view.
        self.column_position = np.random.random_integers(self.column_min,
                                                         self.column_max)
        self.row_position = np.random.random_integers(self.row_min,
                                                      self.row_max)

        self.num_sensors = 2 * self.fov_span ** 2
        self.num_actions = 16
        self.sensors = np.zeros(self.num_sensors)
        self.action = np.zeros(self.num_actions)

        # jump_fraction : float
        #     The fraction of time steps on which the agent jumps to
        #     a random position.
        self.jump_fraction = .05
        self.reward = 0.
        # column_history, row_history : list if ints
        #     A time series of the location (measured in column or row pixels) of the
        #     center of the brain's field of view.
        self.column_history = []
        self.row_history = []
        self.world_visualize_period = 1e6
        self.brain_visualize_period = 1e3
        # print_features : boolean
        #     Indicate whether to visualize each of the features individually.
        #     TODO: re-implement print features
        self.print_features = False


    def step(self, action):
        """
        Advance the world by one time step.

        Parameters
        ----------
        action : array of floats
            The set of action commands to execute.

        Returns
        -------
        self.reward : float
            The amount of reward or punishment given by the world.
        self.sensors : array of floats
            The values of each of the sensors.
        """
        self.timestep += 1
        self.action = action.ravel()
        self.action[np.nonzero(self.action)] = 1.

        # Actions 0-3 move the field of view to a higher-numbered
        # row (downward in the image_data) with varying magnitudes,
        # and actions 4-7 do the opposite.
        # Actions 8-11 move the field of view to a higher-numbered
        # column (rightward in the image_data) with varying magnitudes,
        # and actions 12-15 do the opposite.
        row_step = np.round(action[0] * self.max_step_size / 2 +
                            action[1] * self.max_step_size / 4 +
                            action[2] * self.max_step_size / 8 +
                            action[3] * self.max_step_size / 16 -
                            action[4] * self.max_step_size / 2 -
                            action[5] * self.max_step_size / 4 -
                            action[6] * self.max_step_size / 8 -
                            action[7] * self.max_step_size / 16)
        column_step = np.round(action[8] * self.max_step_size / 2 +
                               action[9] * self.max_step_size / 4 +
                               action[10] * self.max_step_size / 8 +
                               action[11] * self.max_step_size / 16 -
                               action[12] * self.max_step_size / 2 -
                               action[13] * self.max_step_size / 4 -
                               action[14] * self.max_step_size / 8 -
                               action[15] * self.max_step_size / 16)

        row_step = np.round(row_step * (
            1. + np.random.normal(scale=self.noise_magnitude)))
        column_step = np.round(column_step * (
            1. + np.random.normal(scale=self.noise_magnitude)))
        self.row_position = self.row_position + int(row_step)
        self.column_position = self.column_position + int(column_step)

        # Respect the boundaries of the image_data.
        self.row_position = max(self.row_position, self.row_min)
        self.row_position = min(self.row_position, self.row_max)
        self.column_position = max(self.column_position, self.column_min)
        self.column_position = min(self.column_position, self.column_max)

        # At random intervals, jump to a random position in the world.
        if np.random.random_sample() < self.jump_fraction:
            self.column_position = np.random.random_integers(self.column_min,
                                                             self.column_max)
            self.row_position = np.random.random_integers(self.row_min,
                                                          self.row_max)
        self.row_history.append(self.row_position)
        self.column_history.append(self.column_position)

        # Create the sensory input vector.
        fov = self.image_data[int(self.row_position - self.fov_height / 2):
                              int(self.row_position + self.fov_height / 2),
                              int(self.column_position - self.fov_width / 2):
                              int(self.column_position + self.fov_width / 2)]
        # Calculate center surround features for the field of view.
        center_surround_pixels = wtools.center_surround(fov,
                                                        self.fov_span,
                                                        self.fov_span)
        unsplit_sensors = center_surround_pixels.ravel()
        # Center surround values vary between -1 and 1. One means light
        # surrounded by dark, one means dark surrounded by light.
        # Split them each into
        # two sensors, and stack the sets of positive and negative sensors
        # together to complete the sensor array.
        self.sensors = np.concatenate((np.maximum(unsplit_sensors, 0),
                                       np.abs(np.minimum(unsplit_sensors, 0))))

        self.reward = 0
        rewarded_column = (np.abs(self.column_position - self.target_column) <
                           self.reward_region_width / 2)
        rewarded_row = (np.abs(self.row_position - self.target_row) <
                        self.reward_region_width / 2)
        if rewarded_column and rewarded_row:
            self.reward += 1.

        return self.sensors, self.reward


    def visualize_world(self, brain):
        """
        Show what is going on in Becca and in the world.
        """
        if self.print_features:
            projections = brain.get_index_projections()[0]
            wtools.print_pixel_array_features(
                projections,
                self.fov_span ** 2 * 2,
                0,
                self.fov_span,
                self.fov_span,
                world_name=self.name)

        # Periodically display the history and inputs as perceived by Becca.

        print(' '.join(["world is", str(self.timestep), "timesteps old."]))
        fig = plt.figure(11)
        plt.clf()
        plt.plot(self.row_history, 'k.')
        plt.title("Row history")
        plt.xlabel('time step')
        plt.ylabel('position (pixels)')
        fig.show()
        fig.canvas.draw()
        fig = plt.figure(12)
        plt.clf()
        plt.plot(self.column_history, 'k.')
        plt.title("Column history")
        plt.xlabel('time step')
        plt.ylabel('position (pixels)')
        fig.show()
        fig.canvas.draw()

        fig = plt.figure(13)
        clip = (self.sensors[:int(len(self.sensors)/2.)] -
                self.sensors[int(len(self.sensors)/2.):] + 1.) / 2.
        sensed_image = np.reshape(clip, (self.fov_span, self.fov_span))
        plt.gray()
        plt.imshow(sensed_image, interpolation='nearest')
        plt.title("Image sensed")
        fig.show()
        fig.canvas.draw()


if __name__ == "__main__":
    becca.connector.run(World())



"""
Connect a Becca brain to a world and run them.

Command line usage
-----
Test Becca on the grid1D.py world.
    python -m test grid1D
        or
    python -m test 1

Test Becca on the suite of all test worlds.
    python -m test all
        or
    python -m test 0

Profile Becca on the image2D.py world.
    python -m test image2D --profile
        or
    python -m test 9 -p
"""

from __future__ import print_function
import argparse
import cProfile
import pstats
import time

import matplotlib.pyplot as plt
import numpy as np

import becca.connector
# Import the suite of test worlds
from becca_test.grid_1D import World as World_grid_1D
from becca_test.grid_1D_chase import World as World_grid_1D_chase
from becca_test.grid_1D_delay import World as World_grid_1D_delay
from becca_test.grid_1D_ms import World as World_grid_1D_ms
from becca_test.grid_1D_noise import World as World_grid_1D_noise
from becca_test.grid_2D import World as World_grid_2D
from becca_test.grid_2D_dc import World as World_grid_2D_dc
from becca_test.image_1D import World as World_image_1D
from becca_test.image_2D import World as World_image_2D
from becca_test.fruit import World as World_fruit


def suite(lifespan=1e4):
    """
    Run all the worlds in the benchmark and tabulate their performance.
    """
    start_time = time.time()
    performance = []
    performance.append(test_world(World_grid_1D, lifespan=lifespan))
    performance.append(test_world(World_grid_1D_chase, lifespan=lifespan))
    performance.append(test_world(World_grid_1D_delay, lifespan=lifespan))
    performance.append(test_world(World_grid_1D_ms, lifespan=lifespan))
    performance.append(test_world(World_grid_1D_noise, lifespan=lifespan))
    performance.append(test_world(World_grid_2D, lifespan=lifespan))
    performance.append(test_world(World_grid_2D_dc, lifespan=lifespan))
    performance.append(test_world(World_image_1D, lifespan=lifespan))
    performance.append(test_world(World_image_2D, lifespan=lifespan))
    performance.append(test_world(World_fruit, lifespan=lifespan))
    finish_time = time.time()

    # Some tests are harder than others. Weight them accordingly.
    weights = np.array([
        1., # grid1D
        1., # grid1D_chase
        1., # grid1D_delay
        1., # grid1D_ms
        1., # grid1D_noise
        3., # grid2D
        4., # grid2D_dc
        5., # image_1D
        10., # image_2D
        3., # fruit
        ])

    print('Individual test world scores:')
    scores = []
    for score in performance:
        print('    {0:.2}, {1}'.format(score[0], score[1]))
        scores.append(score[0])
    mean_score = np.sum(np.array(scores) * weights) / np.sum(weights)
    print('Weighted test suite score: {0:.2}'.format(mean_score))
    print('Test suite completed in {0:.2} seconds ({1:.2} minutes)'.format(
        finish_time - start_time, (finish_time - start_time) / 60.))

    # Block the program, displaying all plots.
    # When the plot windows are closed, the program closes.
    plt.show()


def test_world(world_class, lifespan=1e4):
    """
    Test the brain's performance on a world.

    Parameters
    ----------
    world_class : World
        The class containing the Becca-compatible world that the
        brain will be receiving sensor and reward information from and
        sending action commands to.
    testing_period : int, optional
        The number of time steps to test the brain
        on the current world.

    Returns
    -------
    performance : float
        The average reward per time step during the testing period.
    world.name : str
        The name of the world that was run.
    """
    start_time = time.time()
    world = world_class(lifespan=lifespan)
    performance = becca.connector.run(world)
    finish_time = time.time()
    delta_time = finish_time - start_time
    print('Performance is: {0:.3}'.format(performance))
    print(world.name, 'ran in {0:.2} seconds ({1:.2} minutes),'.format(
        delta_time, delta_time / 60.))
    print('an average of {0:.2} seconds ({1:.2} ms) per time step.'.format(
        delta_time / lifespan, 1000. * delta_time / lifespan))
    return performance, world.name


def profile(World, lifespan=1e4):
    """
    Profile the brain's performance on the selected world.
    """
    print('Profiling Becca\'s performance...')
    command = 'becca.connector.run(World(lifespan={0}), restore=True)'.format(
        lifespan)
    cProfile.run(command, 'becca_test.profile')
    profile_stats = pstats.Stats('becca_test.profile')
    profile_stats.strip_dirs().sort_stats('time', 'cumulative').print_stats(30)
    print('   View at the command line with')
    print(' > python -m pstats becca_test.profile')


if __name__ == '__main__':
    # Build the command line parser.
    parser = argparse.ArgumentParser(
        description='Test Becca on some toy worlds.')
    parser.add_argument('world', default='all',
                        help=' '.join(['The test world to run.',
                                       'Choose by name or number:',
                                       '1) grid1D,',
                                       '2) grid1D_chase,',
                                       '3) grid1D_delay,',
                                       '4) grid1D_ms,',
                                       '5) grid1D_noise,',
                                       '6) grid2D,',
                                       '7) grid2D_dc,',
                                       '8) image1D,',
                                       '9) image2D,',
                                       '10) fruit,',
                                       '0) all',
                                       'Default value is all.']))
    parser.add_argument(
        '-p', '--profile', action='store_true',
        help="Profile Becca's performance.")
    parser.add_argument(
        '-t', '--lifespan', type=int,
        help='The number of time steps (in thousands) to run the world.')
    args = parser.parse_args()

    if args.world == 'grid1D' or args.world == '1':
        World = World_grid_1D
    elif args.world == 'grid1D_chase' or args.world == '2':
        World = World_grid_1D_chase
    elif args.world == 'grid1D_delay' or args.world == '3':
        World = World_grid_1D_delay
    elif args.world == 'grid1D_ms' or args.world == '4':
        World = World_grid_1D_ms
    elif args.world == 'grid1D_noise' or args.world == '5':
        World = World_grid_1D_noise
    elif args.world == 'grid2D' or args.world == '6':
        World = World_grid_2D
    elif args.world == 'grid2D_dc' or args.world == '7':
        World = World_grid_2D_dc
    elif args.world == 'image1D' or args.world == '8':
        World = World_image_1D
    elif args.world == 'image2D' or args.world == '9':
        World = World_image_2D
    elif args.world == 'fruit' or args.world == '10':
        World = World_fruit
    else:
        args.world = 'all'

    if args.lifespan is None:
        lifespan = 1e4
    else:
        lifespan = args.lifespan * 1000

    if args.world == 'all':
        suite(lifespan=lifespan)
    elif args.profile:
        profile(World, lifespan=lifespan)
    else:
        performance = becca.connector.run(World, restore=True)
        print('performance:', performance)



"""
One-dimensional visual servo task.

This task gives Becca a chance to build a comparatively large number
of sensors into a few informative features. However, due to the
construction of the task, it's not strictly necessary to build
complex features to do well on it.
"""
from __future__ import print_function
import os

import matplotlib.pyplot as plt
import numpy as np

import becca.connector
from becca_test.base_world import World as BaseWorld
import becca_test.world_tools as wtools


class World(BaseWorld):
    """
    One-dimensional visual servo world

    In this world, Becca can direct its gaze left and right
    along a mural. It is rewarded for directing it near the center.
    Optimal performance is a reward of somewhere around .9 per time step.

    Some of this world's attributes are defined in base_world.py.
    The rest are defined below.
    """
    def __init__(self, lifespan=None):
        """
        Set up the world

        Parameters
        ----------
        lifespan : int
            The number of time steps to continue the world.
        """
        BaseWorld.__init__(self, lifespan)
        self.name_long = 'one dimensional visual world'
        self.name = 'image_1D'
        print("Entering", self.name_long)

        # fov_span : int
        #     The world pixelizes its field of view into a superpixel array
        #     that is fov_span X fov_span.
        self.fov_span = 5
        self.num_sensors = 2 * self.fov_span ** 2
        self.num_actions = 8
        # jump_fraction : float
        #     The fraction of time steps on which the agent jumps to
        #     a random position.
        self.jump_fraction = .1
        # step_cost : float
        #     The punishment per position step taken.
        self.step_cost = .1

        self.world_visualize_period = 1e6
        self.brain_visualize_period = 1e3
        # print_features : bool
        #     If True, plot and save visualizations of each of the features
        #     each time the world is visualized,
        #     rendered so that they represent what they mean in this world.
        self.print_features = False

        # Initialize the image to be used as the environment
        module_path = os.path.dirname(os.path.abspath(__file__))
        # image_filename : str
        #     The file name of the image including the relative path.
        self.image_filename = os.path.join(module_path,
                                           'images',
                                           'bar_test.png')
        # data : array of floats
        #     The image, read in and stored as a 2D numpy array.
        self.data = plt.imread(self.image_filename)
        # Convert it to grayscale if it's in color
        if self.data.shape[2] == 3:
            # Collapse the three RGB matrices into one b/w value matrix
            self.data = np.sum(self.data, axis=2) / 3.0
        # Define the size of the field of view, its range of
        # allowable positions, and its initial position
        image_width = self.data.shape[1]
        # max_step_size : int
        #     The largest step size allowed, in pixels in the original image.
        self.max_step_size = image_width / 2
        # target_column : int
        #     The column index that marks the center of the rewarded region.
        self.target_column = image_width / 2
        # reward_region_width : int
        #     The width of the region, in number of columns, within which
        #     the center of the field of view gets rewarded.
        self.reward_region_width = image_width / 8
        # noise_magnitude : float
        #     A scaling factor that drives how much inaccurate each movement
        #     will be.
        self.noise_magnitude = 0.1

        # column_history : list if ints
        #     A time series of the location (measured in column pixels) of the
        #     center of the brain's field of view.
        self.column_history = []
        # fov_height, fov_width : float
        #     The height and width (number of rows) of the field of view,
        #     in pixels.
        self.fov_height = np.min(self.data.shape)
        self.fov_width = self.fov_height
        # column_min, column_max : int
        #     The low and high bounds on where the field of view
        #     can be centered.
        self.column_min = np.ceil(self.fov_width / 2)
        self.column_max = np.floor(self.data.shape[1] - self.column_min)
        # column_position : int
        #     The current location of the center of the field of view.
        self.column_position = np.random.random_integers(self.column_min,
                                                         self.column_max)
        # block_width : int
        #     The width of each superpixel, in number of columns.
        self.block_width = self.fov_width / (self.fov_span + 2)
        self.sensors = np.zeros(self.num_sensors)
        self.action = np.zeros(self.num_actions)
        self.reward = 0.


    def step(self, action):
        """
        Advance the world by one time step

        Parameters
        ----------
        action : array of floats
            The set of action commands to execute.

        Returns
        -------
        self.reward : float
            The amount of reward or punishment given by the world.
        self.sensors : array of floats
            The values of each of the sensors.
        """
        self.timestep += 1
        self.action = action.ravel()
        self.action[np.nonzero(self.action)] = 1.

        # Actions 0-3 move the field of view to a higher-numbered row
        # (downward in the image) with varying magnitudes, and
        # actions 4-7 do the opposite.
        raw_col_step = (self.action[0] * self.max_step_size / 2 +
                        self.action[1] * self.max_step_size / 4 +
                        self.action[2] * self.max_step_size / 8 +
                        self.action[3] * self.max_step_size / 16 -
                        self.action[4] * self.max_step_size / 2 -
                        self.action[5] * self.max_step_size / 4 -
                        self.action[6] * self.max_step_size / 8 -
                        self.action[7] * self.max_step_size / 16)
        noise_factor = (
            self.noise_magnitude * np.random.random_sample() * 2.0 -
            self.noise_magnitude * np.random.random_sample() * 2.0 + 1.)
        column_step = int(raw_col_step * noise_factor)
        self.column_position = self.column_position + column_step
        self.column_position = max(self.column_position, self.column_min)
        self.column_position = min(self.column_position, self.column_max)
        self.column_history.append(self.column_position)

        # At random intervals, jump to a random position in the world.
        if np.random.random_sample() < self.jump_fraction:
            self.column_position = np.random.random_integers(self.column_min,
                                                             self.column_max)
        # Create the sensory input vector.
        fov = self.data[:, int(self.column_position - self.fov_width / 2):
                        int(self.column_position + self.fov_width / 2)]
        # Calculate center surround features for the image.
        center_surround_pixels = wtools.center_surround(fov,
                                                        self.fov_span,
                                                        self.fov_span)
        unsplit_sensors = center_surround_pixels.ravel()
        # These can be positive or negative, so split them into two
        # sets of sensors--one for the positive values and one for the
        # negative ones. Then stack them together for one big sensor array.
        self.sensors = np.concatenate((np.maximum(unsplit_sensors, 0),
                                       np.abs(np.minimum(unsplit_sensors, 0))))

        # Calculate the reward.
        self.reward = 0
        if (np.abs(self.column_position - self.target_column) <
                self.reward_region_width / 2.0):
            self.reward += 1.
        self.reward -= (np.abs(column_step) /
                        self.max_step_size * self.step_cost)
        return self.sensors, self.reward


    def visualize_world(self, brain):
        """
        Show what's going on in the world.
        """
        if self.print_features:
            projections = brain.cortex.get_index_projections(to_screen=True)[0]
            wtools.print_pixel_array_features(
                projections,
                self.fov_span ** 2 * 2,
                0,
                self.fov_span, self.fov_span,
                world_name=self.name)

        # Periodically show the state history and inputs as perceived by Becca.
        print(''.join(["world is ", str(self.timestep), " timesteps old"]))
        fig = plt.figure(11)
        plt.clf()
        plt.plot(self.column_history, 'k.', alpha=.2)
        plt.title(''.join(['Column history for ', self.name]))
        plt.xlabel('time step')
        plt.ylabel('position (pixels)')
        fig.show()
        fig.canvas.draw()

        fig = plt.figure(12)
        sensed_image = np.reshape(
            0.5 * (self.sensors[:len(self.sensors)/2] -
                   self.sensors[len(self.sensors)/2:] + 1),
            (self.fov_span, self.fov_span))
        plt.gray()
        plt.imshow(sensed_image, interpolation='nearest')
        plt.title("Image sensed")
        fig.show()
        fig.canvas.draw()


if __name__ == "__main__":
    becca.connector.run(World())



"""
A few functions that are useful to multiple worlds
"""
from __future__ import print_function
import os

import matplotlib.pyplot as plt
import numpy as np

import becca.tools as tools


def center_surround(fov, fov_horz_span, fov_vert_span, verbose=False):
    """
    Convert a 2D array of b/w pixel values to center-surround

    Parameters
    ----------
    fov : 2D array of floats
         Pixel values from the field of view.
    fov_horz_span : int
        Desired number of center-surround superpixel columns.
    fov_vert_span: int
        Desired number of center-surround superpixel rows.
    verbose : bool
        If True, print more information to the console.

    Returns
    -------
    center_surround_pixels :  2D array of floats
        The center surround values corresponding to the inputs.
    """
    fov_height = fov.shape[0]
    fov_width = fov.shape[1]
    block_width = float(fov_width) / float(fov_horz_span + 2)
    block_height = float(fov_height) / float(fov_vert_span + 2)
    super_pixels = np.zeros((fov_vert_span + 2, fov_horz_span + 2))
    center_surround_pixels = np.zeros((fov_vert_span, fov_horz_span))
    # Create the superpixels by averaging pixel blocks
    for row in range(fov_vert_span + 2):
        for col in range(fov_horz_span + 2):
            super_pixels[row][col] = np.mean(
                fov[int(float(row) * block_height):
                    int(float(row + 1) * block_height),
                    int(float(col) * block_width) :
                    int(float(col + 1) * block_width)])
    for row in range(fov_vert_span):
        for col in range(fov_horz_span):
            # Calculate a center-surround value that represents
            # the difference between the pixel and its surroundings.
            # Weight the N, S, E, and W pixels by 1/6 and
            # the NW, NE, SW, and SE pixels by 1/12, and
            # subtract from the center.
            center_surround_pixels[row][col] = (
                super_pixels[row + 1][col + 1] -
                super_pixels[row][col + 1] / 6 -
                super_pixels[row + 2][col + 1] / 6 -
                super_pixels[row + 1][col] / 6 -
                super_pixels[row + 1][col + 2] / 6 -
                super_pixels[row][col] / 12 -
                super_pixels[row + 2][col] / 12 -
                super_pixels[row][col + 2] / 12 -
                super_pixels[row + 2][col + 2] / 12)
    if verbose:
        # Display the field of view clipped from the original image
        plt.figure("fov")
        plt.gray()
        img = plt.imshow(fov)
        img.set_interpolation('nearest')
        plt.title("field of view")
        plt.draw()

        # Display the pixelized version, a.k.a. superpixels
        plt.figure("super_pixels")
        plt.gray()
        img = plt.imshow(super_pixels)
        img.set_interpolation('nearest')
        plt.title("super pixels")
        plt.draw()

        # Display the center-surround filtered superpixels
        plt.figure("center_surround_pixels")
        plt.gray()
        img = plt.imshow(center_surround_pixels)
        img.set_interpolation('nearest')
        plt.title("center surround pixels")
        plt.draw()
    return center_surround_pixels


def print_pixel_array_features(projections,
                               num_pixels_x2,
                               start_index,
                               fov_horz_span,
                               fov_vert_span,
                               directory='log',
                               world_name='',
                               interp='nearest'):
    """
    Interpret an array of center-surround pixels as an image.

    Parameters
    ----------
    directory : str
       The directory into which the feature images will be saved.
       Default is 'log'.
    fov_horz_span, fov_vert_span : int
        The number of pixels in the horizontal (columns) and vertical (rows)
        directions.
    interp : str
        The method of interpolation that matplotlib will use when creating
        the image. Default is 'nearest'.
    num_pixels_x2 : int
        Twice the number of center-surround superpixels.
    projections : list of list of array of floats
        This is the set of all the projections of all the features from
        all the ``ZipTie``s onto sensors.
    start_index : int
        Which index in the projection arrays marks the beginning of
        the center-surround sensors.
    world_name : str
        A base name for the image filenames, associated with the world.
    """
    num_levels = len(projections)
    for level_index in range(num_levels):
        for feature_index in range(len(projections[level_index])):
            plt.close(99)
            feature_fig = plt.figure(num=99)

            # Get the pixel array for the projection image.
            projection_image = (visualize_pixel_array_feature(
                projections[level_index][feature_index]
                [start_index:start_index + num_pixels_x2],
                fov_horz_span, fov_vert_span, array_only=True))
            rect = (0., 0., 1., 1.)
            axes = feature_fig.add_axes(rect)
            plt.gray()
            axes.imshow(projection_image,
                        interpolation=interp, vmin=0., vmax=1.)

            # Create a plot of individual features.
            filename = '_'.join(['level', str(level_index).zfill(2),
                                 'feature', str(feature_index).zfill(4),
                                 world_name, 'image.png'])
            full_filename = os.path.join(directory, filename)
            plt.title(filename)
            plt.savefig(full_filename, format='png')


def visualize_pixel_array_feature(sensors,
                                  fov_horz_span=None,
                                  fov_vert_span=None,
                                  level_index=-1,
                                  feature_index=-1,
                                  world_name=None,
                                  save_png=False,
                                  filename='log/feature',
                                  array_only=False):
    """
    Show a visual approximation of an array of center-surround sensorss.

    Parameters
    ----------
    array_only : bool
        If True, calculate but do not plot the pixel values of the image.
        Default is False.
    feature_index : int
        The index of the feature in its ``ZipTie``.
    filename : str
        The base filename under which each feature visualization image
        is saved. The default is 'log/feature'.
    fov_horz_span : int
        Desired number of center-surround superpixel columns.
    fov_vert_span: int
        Desired number of center-surround superpixel rows.
    level_index : int
        The index of the ``ZipTie`` level from which the feature is taken.
    save_png : bool
        If True, save a copy of the visualization as a png. Default is False.
    sensors : array of floats
        This assumes that sensors are arranged as a set of flattened
        superpixel brightness sernsors concatenated with the complementary
        set of flattened superpixel darkness sensors (1. - brightness).
    world_name : str
        The name of the world that generated the features.

    Returns
    -------
    feature_pixels : 2D array of floats
        If ``array_only``, return the array of image values.
    """
    # Calculate the number of pixels that span the field of view
    if fov_horz_span is None:
        n_pixels = sensors.shape[0] / 2.
        fov_horz_span = np.sqrt(n_pixels)
        fov_vert_span = np.sqrt(n_pixels)
    else:
        n_pixels = fov_horz_span * fov_vert_span

    # Maximize contrast
    sensors *= 1. / (np.max(sensors) + tools.epsilon)

    # Calculate the visualization image pixel values.
    pixel_values = ((sensors[0:n_pixels] -
                     sensors[n_pixels:2 * n_pixels]) + 1.0) / 2.0
    feature_pixels = pixel_values.reshape(fov_vert_span, fov_horz_span)

    if array_only:
        return feature_pixels
    else:
    # Initialize the and plot the figure.
        level_str = str(level_index).zfill(2)
        feature_str = str(feature_index).zfill(3)
        fig_title = ' '.join(('Level', level_str, 'Feature', feature_str,
                              'from', world_name))
        fig_name = ' '.join(('Features from ', world_name))
        fig = plt.figure(tools.str_to_int(fig_name))
        fig.clf()
        plt.gray()
        fig.add_axes(0., 0., 1., 1., frame_on=False)
        plt.imshow(feature_pixels, vmin=0.0, vmax=1.0,
                   interpolation='nearest')
        plt.title(fig_title)

        # Save the image file.
        if save_png:
            filename = ''.join([filename, '_', world_name, '_',
                                level_str, '_', feature_str, '.png'])
            fig.savefig(filename, format='png')

        # Force an update of the screen display.
        fig.show()
        fig.canvas.draw()


def resample2D(array, num_rows, num_cols):
    """
    Resample a 2D array to get one that has num_rows and num_cols.

    Use and approximate nearest neighbor method to resample to the
    pixel on the next lower row and column..

    Parameters
    ----------
    array : 2D array of floats
        The array to resample from.
    num_cols, num_rows : ints
        The number of rows and columns to include in the
        evenly-spaced grid resampling.

    Returns
    -------
    resampled_array : 2D array of floats
        The resampled version of the array with the appropriate dimensions.
    """
    rows = (np.linspace(0., .9999999, num_rows) *
            array.shape[0]).astype(np.int)
    cols = (np.linspace(0., .9999999, num_cols) *
            array.shape[1]).astype(np.int)

    if len(array.shape) == 2:
        resampled_array = array[rows, :]
        resampled_array = resampled_array[:, cols]
    if len(array.shape) == 3:
        resampled_array = array[rows, :, :]
        resampled_array = resampled_array[:, cols, :]
    return resampled_array



"""
A set of simple worlds that can be used to evaluate Becca agents

These are toy worlds, but they are diverse.
Their purpose is to push the limits of typical learning agents
is as simple a manner as possible.
"""



"""
The base world on which all the other worlds are based.
"""
from __future__ import print_function
import numpy as np


class World(object):
    """
    The base class for creating a new world.
    """
    def __init__(self, lifespan=None):
        """
        Initialize a new world with some benign default values.

        Parameters
        ----------
        lifespan : int or None
            The number of time steps that the world will be
            allowed to continue.
        """
        # lifespan : float
        #     The number of time steps for which the world should continue
        #     to exist.
        if lifespan is None:
            self.lifespan = 10 ** 5
        else:
            self.lifespan = lifespan
        # timestep : int
        #     The number of time steps that the world has already been through.
        #     Starting at -1 allows for an intialization pass.
        self.timestep = -1
        # world_visualization_period : int
        #     How often to turn the world into a picture.
        self.world_visualize_period = 1e6
        # brain_visualization_period : int
        #     How often to turn the brain's state into a picture.
        self.brain_visualize_period = 1e4
        # name : String
        #     The name of the world.
        self.name = 'abstract base world'
        # num_actions, num_sensors : int
        #     The number of actions and sensors, respectively.
        #     These will likely be overridden in any subclass
        self.num_sensors = 0
        self.num_actions = 0
        # sensors, actions : array of floats
        #     The arrays that represent the full set of sensed observations
        #     and intended actions, updated at each time step.
        self.sensors = np.zeros(self.num_sensors)
        self.actions = np.zeros(self.num_actions)
        # reward : float
        #     The feedback signal on the goodness of an brain's experience.
        #     0 is neutral.
        #     1 is very, very good.
        #     -1 is very, very bad.
        self.reward = 0


    def step(self, actions):
        """
        Take a time step through an empty world that does nothing.

        Parameters
        ----------
        actions : array of floats
            The set of actions that the world can be expected to execute.

        Returns
        -------
        sensors : array of floats
            The current values of each of those sensors in the world.
        reward : float
            The current reward provided by the world.
        """
        self.timestep += 1
        self.sensors = np.zeros(self.num_sensors)
        self.actions = np.zeros(self.num_actions)
        self.reward = 0
        return self.sensors, self.reward


    def is_alive(self):
        """
        Check whether the world is alive.

        Once more than lifespan time steps have been completed,
        stop running.

        Returns
        -------
        If False, the world has come to an end.
        """
        return self.timestep < self.lifespan


    def visualize(self, brain):
        """
        Let the world show Becca's internal state as well as its own.

        Parameters
        ----------
        brain : Brain
            A copy of the Brain, provided to the world so that the
            world can interpret and visualize it in the context of the
            world.
        """
        if (self.timestep % self.world_visualize_period) == 0:
            self.visualize_world(brain)
        if (self.timestep % self.brain_visualize_period) == 0:
            brain.visualize()


    def visualize_world(self, brain):
        """
        Show the user the state of the world.
        """
        print('{0} is {1} time steps old.'.format(self.name, self.timestep))
        print('The brain is {0} time steps old.'.format(brain.timestep))




"""
One-dimensional grid task.

This task tests a brain's ability to choose an appropriate action.
It is straightforward. Reward and punishment is clear and immediate.
There is only one reward state and it can be reached in a single
step.

Usage
To run this world standalone from the command line

    python -m grid_1D

"""
from __future__ import print_function
import numpy as np

import becca.connector
from becca_test.base_world import World as BaseWorld

class World(BaseWorld):
    """
    One-dimensional grid world.

    In this task, the brain steps forward and backward along
    a nine-position line. The fourth position is rewarded and
    the ninth position is punished. There is also a slight
    punishment for effort expended in taking actions.
    Occasionally the brain will get
    involuntarily bumped to a random position on the line.
    This is intended to be a simple-as-possible
    task for troubleshooting Becca.
    Optimal performance is a reward of about 90 per time step.

    Most of this world's attributes are defined in base_world.py.
    The few that aren't are defined below.
    """
    def __init__(self, lifespan=None):
        """
        Initialize the world.

        Parameters
        ----------
        lifespan : int
            The number of time steps to continue the world.
        """
        BaseWorld.__init__(self, lifespan)
        self.name = 'grid_1D'
        self.name_long = 'one dimensional grid world'
        print("Entering", self.name_long)

        self.num_sensors = 9
        self.num_actions = 8
        self.action = np.zeros(self.num_actions)
        self.energy = 0.
        # energy_cost : float
        #     The punishment per position step taken.
        self.energy_cost = 1. / 100.
        # world_state : float
        #     The actual position of the agent in the world.
        #     This can be fractional.
        self.world_state = 0
        # simple_state : int
        #     The nearest integer position of the agent in the world.
        self.simple_state = 0
        # jump_fraction : float
        #     The fraction of time steps on which the agent jumps to
        #     a random position.
        self.jump_fraction = 0.1

        self.world_visualize_period = 1e6
        self.brain_visualize_period = 1e3


    def step(self, action):
        """
        Advance the world one time step.

        Parameters
        ----------
        action : array of floats
            The set of action commands to execute.

        Returns
        -------
        reward : float
            The amount of reward or punishment given by the world.
        sensors : array of floats
            The values of each of the sensors.
        """
        self.action = action
        self.action = np.round(self.action)
        self.timestep += 1

        # Find the step size as combinations of the action commands
        #     action[i]     result
        #            0      1 step to the right
        #            1      2 steps to the right
        #            2      3 steps to the right
        #            3      4 steps to the right
        #            4      1 step to the left
        #            5      2 steps to the left
        #            6      3 steps to the left
        #            7      4 steps to the left
        step_size = (self.action[0] +
                     self.action[1] * 2 +
                     self.action[2] * 3 +
                     self.action[3] * 4 -
                     self.action[4] -
                     self.action[5] * 2 -
                     self.action[6] * 3 -
                     self.action[7] * 4)
        # Action cost is an approximation of metabolic energy.
        # Action cost is proportional to the number of steps taken.
        self.energy = (self.action[0] +
                       self.action[1] * 2 +
                       self.action[2] * 3 +
                       self.action[3] * 4 +
                       self.action[4] +
                       self.action[5] * 2 +
                       self.action[6] * 3 +
                       self.action[7] * 4)

        self.world_state = self.world_state + step_size

        # At random intervals, jump to a random position in the world.
        if np.random.random_sample() < self.jump_fraction:
            self.world_state = self.num_sensors * np.random.random_sample()

        # Ensure that the world state falls between 0 and 9.
        self.world_state -= self.num_sensors * np.floor_divide(
            self.world_state, self.num_sensors)
        self.simple_state = int(np.floor(self.world_state))
        if self.simple_state == 9:
            self.simple_state = 0

        # Represent the presence or absence of the current position in the bin.
        sensors = np.zeros(self.num_sensors)
        sensors[self.simple_state] = 1
        reward = self.assign_reward(sensors)

        return sensors, reward


    def assign_reward(self, sensors):
        """
        Calculate the total reward corresponding to the current state

        Parameters
        ----------
        sensors : array of floats
            The current sensor values.

        Returns
        -------
        reward : float
            The reward associated the set of input sensors.
        """
        reward = 0.
        reward -= sensors[8]
        reward += sensors[3]
        # Punish actions just a little
        reward -= self.energy  * self.energy_cost
        reward = np.maximum(reward, -1.)

        return reward


    def visualize_world(self, brain):
        """
        Show what's going on in the world.
        """
        state_image = ['.'] * (self.num_sensors + self.num_actions + 2)
        state_image[self.simple_state] = 'O'
        state_image[self.num_sensors:self.num_sensors + 2] = '||'
        action_index = np.where(self.action > 0.1)[0]
        if action_index.size > 0:
            for i in range(action_index.size):
                state_image[self.num_sensors + 2 + action_index[i]] = 'x'
        print(''.join(state_image))


if __name__ == "__main__":
    becca.connector.run(World())
